; ==============================================================================
; file: "source/system/kernel.asm"
; use: to be compiled into "system/kernel.bin"
; ==============================================================================

use32
org 0x20000

glaux_define_calls:                     ; NEVER CHANGE!

; TODO for 0.1.1: check and document all calls and fill in the gaps
; TODO for 0.2.0: push and pop all altered variables in system calls

    .entry:
        jmp kernel                      ; 0x00020000
        
    .display:
        jmp display_fill                ; 0x00020005
        jmp display_update              ; 0x0002000A
        jmp display_paste_str           ; 0x0002000F
        jmp display_paste_rect          ; 0x00020014
        jmp display_paste_char          ; 0x00020019
        jmp display_scroll              ; 0x0002001E
        times 256 - ($ - .entry) db 0

    .string:
        jmp string_compare              ; 0x00020100
        times 256 - ($ - .string) db 0

    .keybd:
        jmp keybd_get_scan              ; 0x00020200
        jmp keybd_conv_latin            ; 0x00020205
        jmp keybd_input_str             ; 0x0002020A
        times 512 - ($ - .keybd) db 0
    
    .ehci:
        ; it is dangerous to have these calls accessible to user programs
    
    .pci:
        ; it is dangerous to have these calls accessible to user programs
    
    .math:
        jmp math_int_to_hex             ; 0x00020400
        times 256 - ($ - .math) db 0
    
    .ext2:
        jmp ext2_read_file              ; 0x00020500
        times 256 - ($ - .ext2) db 0

; ==============================================================================

glaux_include_calls_vars:
    %include "source/system/kernel/display.inc" ; display calls
    %include "source/system/kernel/string.inc"  ; string calls
    %include "source/system/kernel/keybd.inc"   ; keyboard calls
    %include "source/system/kernel/ehci.inc"    ; ehci calls
    %include "source/system/kernel/pci.inc"     ; pci calls
    %include "source/system/kernel/math.inc"    ; math calls
    %include "source/system/kernel/ext2.inc"    ; ext2 calls
    %include "source/variables.inc"             ; variables used for above calls

; ==============================================================================

kernel:
    
    .clear_display:
        mov dword [color_1], 0          ; we make screen black
        mov dword [color_2], 0xCCFF99   ; and we prepare green letters
        call display_fill
        call display_update

    .init_ehci_ctrller:
        call ehci_init_ctrller          ; we init the ehci ctrller
        call ehci_init_dev0             ; and the boot device
    
    .load_font:
        mov dword [string_1], font_path         ; we load the default font
        mov dword [ram_address], font_buffer    ; at 0x00200000
        call ext2_read_file
    
    .load_keybd_layout:
        mov dword [string_1], layout_path       ; we load the keyboard layout
        mov dword [ram_address], keybd_layout   ; at 0x00100000
        call ext2_read_file
        
    .go_to_start_of_screen:
        mov dword [pixel_horiz], 0      ; we need to go to the top left border
        mov dword [pixel_vert], 0       ; so the prompt starts properly
    
    .reset_char:
        mov dword [char_unicode], 0     ; we start printing at char unicode 0
    
    .print_char:
        call display_paste_char         ; we print current char
        add dword [char_unicode], 1     ; we go to next char
        cmp dword [char_unicode], 128   ; but we showcase only first 128
        je .add_line
        add dword [pixel_horiz], 8      ; we go 8 pixels right
        cmp dword [pixel_horiz], 640    ; if right border, we need to go down
        jb .print_char
        mov dword [pixel_horiz], 0      ; we go to left border
        add dword [pixel_vert], 16      ; and 16 pixels down
        jmp .print_char
        
    .get_next_str:
        mov dword [string_1], prompt    ; we paste prompt string
        call display_paste_str
        call display_update
        mov dword [string_1], cmd       ; we get string from keyboard
        call keybd_input_str
    
    .add_line:
        mov dword [pixel_horiz], 0      ; we go to left border
        add dword [pixel_vert], 16      ; and 16 pixels down
        cmp dword [pixel_vert], 464     ; if this line does not fit
        jbe .get_next_str               ; we need to scroll
    
    .scroll:
        call display_scroll
        call display_update
        jmp .get_next_str

prompt      db  ">",0," ",0,0,0                         ; prompt for terminal
font_path   db "/dev0/system/fonts/glaux-mono.sys",0    ; ascii path for font
layout_path db "/dev0/system/layout.sys",0              ; ascii path for layout

times 16384 - ($ - $$) db 0             ; pad with zeros until 16 kilobytes

cmd:                                    ; we store the command after the kernel
