; ==============================================================================
; file: "source/system/kernel/ehci.inc"
; use: to be included in "source/system/kernel.asm"
; ==============================================================================

ehci_find_ctrller:      

    .return:
        ret

; ==============================================================================

ehci_init_ctrller:

; ----------------------------------------------
; description: inits ehci controller
; input: dword [ehci_bus] \
;        dword [ehci_dev]  > ehci address on pci
;        dword [ehci_fun] /
; output: {none}
; ----------------------------------------------

    .set_ehci_address:
        mov eax, dword [ehci_bus]       ; we set ehci address on pci bus
        mov dword [pci_bus], eax
        mov eax, dword [ehci_dev]
        mov dword [pci_dev], eax
        mov eax, dword [ehci_fun]
        mov dword [pci_fun], eax
        call pci_set_config_address     ; pci_address = ehci bus:dev:fun enabled

    .enable_bus_master_and_memory_space:
        mov dword [pci_offset], 0x04    ; we read dword from offset 4
        call pci_read_dword
        or dword [pci_data], 0x06       ; we set bus master enable
        call pci_write_dword            ; and memory space enable bits
    
    .get_base_address:
        mov dword [pci_offset], 0x10    ; offset for usb base address is 0x10
        call pci_read_dword
        mov eax, dword [pci_data]       ; bus address is bits 31:8
        and eax, 0xFFFFFF00             ; so we have to mask the rest
        mov dword [ehci_base], eax
    
    .get_opbase_address:
        mov esi, dword [ehci_base]       ; we get byte at ehci base address
        movzx eax, byte [esi]
        add eax, esi                     ; we add ehci base address to it
        mov dword [ehci_opbase], eax     ; we save ehci operation base address
    
    .get_eecp_offset:
        movzx eax, word [esi + 8]        ; eecp offset is given in third dword
        shr eax, 8                      ; but we need to shift it one byte right
        mov dword [ehci_eecp], eax       ; because we read dwords and not bytes
    
    .check_valid_eecp:
        cmp eax, 0x40                    ; if eecp pointer < 40
        jb .stop_ctrller                 ; extended capabilities unsupported
    
    .check_bios_control:
        mov dword [pci_offset], eax      ; we read from eecp
        call pci_read_dword
        test dword [pci_data], 0x10000   ; if bit 16 is set, bios owns ehci
        jz .stop_ctrller                 ; else, we skip this step
    
    .get_control:
        or dword [pci_data], 0x1000000   ; we set bit 24 to get ownership
        call pci_write_dword             ; from the bios

    .pause_1:
        mov ecx, 0x8000000               ; we need to wait and we have not
                                         ; a time_wait function, TODO: make it
    .pause_2:
        sub ecx, 1
        jnz .pause_2
    
    .stop_ctrller:
        mov esi, dword [ehci_opbase]     ; to stop the ctrller we need
        and dword [esi], 0xFFFFFFFE      ; to clear bit 0 at first dword
                                         ; of operational registers
    .check_halted:
        test dword [esi + 4], 0x1000     ; to check if the ctrller is halted
        jz .check_halted                 ; we need to read bit 16
                                         ; of second dword of operational regs
    .reset_ctrller:
        mov esi, dword [ehci_opbase]     ; to reset it we need to set bit 1
        or dword [esi], 0x02             ; of first dword of operational regs
    
    .check_reset:
        test dword [esi], 0x02           ; we check if the bit is actually set
        jnz .check_reset                 ; if not, we need to wait until it is
    
    .disable_ints:
        mov esi, dword [ehci_opbase]     ; all interrupts are disabled
        mov dword [esi + 8], 0
    
    .clear_control_data_struct_segment:
        mov dword [esi + 16], 0          ; clear 63:32 bits of 64 bit addressing
    
    .set_QH_link_pointer:
        mov dword [ehci_control_QH], ehci_control_QH
        or dword [ehci_control_QH], 2    ; horiz link pointer points to itself
    
    .set_QH_endpoint_characteristics:
        mov dword [ehci_control_QH + 4], 0x5200E000 ; we set endpoint
                                                    ; characteristics
    .set_QH_endpoint_capabilities:
        mov dword [ehci_control_QH + 8], 0x40000000 ; we set one transaction
                                                    ; per microframe
    .set_current_TD_pointer:
        mov dword [ehci_control_QH + 12], 0 ; written actually by controller
    
    .set_TD_overlay_area:
        mov dword [ehci_control_QH + 16], 1 ; next td pointer
        mov dword [ehci_control_QH + 20], 0
        mov dword [ehci_control_QH + 24], 0
        mov dword [ehci_control_QH + 28], 0
        mov dword [ehci_control_QH + 32], 0
        mov dword [ehci_control_QH + 36], 0
        mov dword [ehci_control_QH + 40], 0
        mov dword [ehci_control_QH + 44], 0
        
        mov dword [ehci_control_QH + 48], 0 ; extended buffer pointer pages
        mov dword [ehci_control_QH + 52], 0
        mov dword [ehci_control_QH + 56], 0
        mov dword [ehci_control_QH + 60], 0
        mov dword [ehci_control_QH + 64], 0
    
    .save_address_of_QH:
        mov esi, dword [ehci_opbase]          ; address of queue head
        mov dword [esi + 24], ehci_control_QH ; needs to be saved
    
    .set_config_flag:
        or dword [esi + 64], 1          ; all ports belong to ehci
        
    .pause_3:
        mov ecx, 0x10000000             ; we need to wait
    
    .pause_4:
        sub ecx, 1
        jnz .pause_4
    
    .enable_async_list:
        mov esi, dword [ehci_opbase]    ; async list is enabled
        or dword [esi], 0x20
    
    .start_ctrller:
        or dword [esi], 1               ; bit 0 must be set for ctrller to start
    
    .pause_5:
        mov ecx, 0x10000000             ; we need to wait again
    
    .pause_6:
        sub ecx, 1
        jnz .pause_6
    
    .clear_toggle_vars:
        mov dword [ehci_in_toggle], 0
        mov dword [ehci_out_toggle], 0
        mov dword [ehci_control_toggle], 0

    .return:
        ret

; ==============================================================================

ehci_make_run_TD:

    .prepare:
        mov esi, ehci_TD_chain
        mov edi, ehci_TD_chain + 64
        mov eax, dword [ehci_TD_buffer_size]
        mov dword [.remaining], eax

    .modify_QH_endpoint:
        mov eax, dword [ehci_control_QH + 4]
        and eax, 0FFFFF0FFh
        mov ebx, dword [ehci_TD_endpoint]
        shl ebx, 8
        or eax, ebx
    
    .modify_QH_address:
        and eax, 0FFFFFF80h
        or eax, dword [ehci_TD_bus_address]
        mov dword [ehci_control_QH + 4], eax
    
    ; ----------------------
    ; first and second dword
    ; ----------------------
    
    .start_loop:
        cmp dword [.remaining], 512
        jbe .terminate_chain
    
    .link_to_next_TD:
        mov dword [esi], edi
        mov dword [esi + 4], 1
        jmp .get_toggle
    
    .terminate_chain:
        mov dword [esi], 1
        mov dword [esi + 4], 1
    
    ; -----------
    ; third dword
    ; -----------
    
    .get_toggle:
        mov ebx, dword [ehci_TD_toggle_pointer]
        mov eax, dword [ebx]
        shl eax, 31
        xor dword [ebx], 1
    
    .get_bytes_to_transfer:
        cmp dword [.remaining], 512
        jbe .set_remaining
    
    .set_512:
        mov ebx, 512
        shl ebx, 16
        or eax, ebx
        jmp .set_error_counter
    
    .set_remaining:
        mov ebx, dword [.remaining]
        shl ebx, 16
        or eax, ebx
    
    .set_error_counter:
        or eax, 0C00h
    
    .get_packet_id:
        mov ebx, dword [ehci_TD_packet_id]
        shl ebx, 8
        or eax, ebx
    
    .set_active_status:
        or eax, 80h
    
    .write_third_dword_to_TD:
        mov dword [esi + 8], eax
    
    ; -----------
    ; next dwords
    ; -----------
    
    .set_buffer_pointer_pages:
        mov eax, dword [ehci_TD_buffer_pointer]
        mov dword [esi + 12], eax
        mov dword [esi + 16], 0
        mov dword [esi + 20], 0
        mov dword [esi + 24], 0
        mov dword [esi + 28], 0
        ; and high 32 bits
        mov dword [esi + 32], 0
        mov dword [esi + 36], 0
        mov dword [esi + 40], 0
        mov dword [esi + 44], 0
        mov dword [esi + 48], 0
    
    .compare_end:
        cmp dword [.remaining], 512
        jbe .run_TD
    
    .go_to_next_TD:
        add esi, 64
        add edi, 64
        sub dword [.remaining], 512
        add dword [ehci_TD_buffer_pointer], 512
        jmp .start_loop
    
    .run_TD:
        mov dword [ehci_control_QH + 16], ehci_TD_chain
    
    .pause_1:
        mov ecx, 04000000h
    
    .pause_2:
        sub ecx, 1
        jz .return
    
    .check_for_halted:
        test dword [esi + 8], 40h
        jnz .pause_2
    
    .check_for_inactive:
        test dword [esi + 8], 80h
        jnz .pause_2
    
    .return:
        ret
    
    .remaining dd 0

; ==============================================================================

ehci_init_dev0:
    
    .set_port_0:
        mov dword [ehci_port_number], 0
    
    .check_port_connection:
        call ehci_check_connected_port
        jz .go_to_next_port
    
    .check_low_speed:
        call ehci_check_low_speed_port
        je .go_to_next_port
    
    .reset_port:
        call ehci_reset_port
    
    .get_device_descriptor:
        call ehci_get_device_descriptor
        call ehci_get_device_descriptor
    
    .get_config_descriptor:
        mov dword [ehci_TD_buffer_size], 9
        call ehci_get_config_descriptor
    
    .get_remaining_descriptors:
        movzx eax, byte [ehci_config_descriptor + 2]
        mov dword [ehci_TD_buffer_size], eax
        call ehci_get_config_descriptor
    
    .check_mass_storage:
        cmp byte [ehci_config_descriptor + 14], 08h
        jne .go_to_next_port
    
    .check_scsi:
        cmp byte [ehci_config_descriptor + 15], 06h
        jne .go_to_next_port
    
    .check_bulk_only:
        cmp byte [ehci_config_descriptor + 16], 50h
        jne .go_to_next_port
    
    .get_endpoint_1:
        mov al, byte [ehci_config_descriptor + 20]
        test al, 80h
        jnz .save_endpoint_in_1
    
    .save_endpoint_out_1:
        mov byte [ehci_out_endpoint], al
        and dword [ehci_out_endpoint], 0Fh
        jmp .get_endpoint_2
    
    .save_endpoint_in_1:
        mov byte [ehci_in_endpoint], al
        and dword [ehci_in_endpoint], 0Fh
    
    .get_endpoint_2:
        mov al, byte [ehci_config_descriptor + 27]
        test al, 80h
        jnz .save_endpoint_in_2
        
    .save_endpoint_out_2:
        mov byte [ehci_out_endpoint], al
        and dword [ehci_out_endpoint], 0Fh
        ; jmp .return
        jmp .set_address
    
    .save_endpoint_in_2:
        mov byte [ehci_in_endpoint], al
        and dword [ehci_in_endpoint], 0Fh
    
    ; until here ok
    
    .set_address:
        mov dword [ehci_bus_address], 1
        call ehci_set_address
    
    .set_config:
        call ehci_set_config
    
    .reset_dev0_toggle:
        mov dword [ehci_in_toggle], 0
        mov dword [ehci_out_toggle], 0
    
        call ehci_inquiry
        call ehci_test_unit
        call ehci_request_sense
        call ehci_test_unit
        call ehci_request_sense
        
        call ehci_read_capacity
        mov eax, dword [ehci_capacity]
        bswap eax
        mov dword [ehci_capacity], eax
        mov eax, dword [ehci_block_size]
        bswap eax
        mov dword [ehci_block_size], eax
        
        jmp .return
    
    .go_to_next_port:
        add dword [ehci_port_number], 1
        cmp dword [ehci_port_number], 4
        jb .check_port_connection
    
    .return:
        ret

; ==============================================================================

ehci_check_connected_port:

    .get_port_number:
        mov eax, dword [ehci_port_number]
    
    .check_connected_port:
        mov esi, dword [ehci_opbase]
        test dword [esi + 44h + eax*4], 1

    .return:
        ret

; ==============================================================================

ehci_check_low_speed_port:
    
    .get_port_number:
        mov eax, dword [ehci_port_number]
    
    .check_low_speed:
        mov esi, dword [ehci_opbase]
        mov ebx, dword [esi + 44h + eax*4]
        shr ebx, 10
        and ebx, 3
        cmp ebx, 1
    
    .return:
        ret
        
; ==============================================================================

ehci_reset_port:
    
    .get_port_number:
        mov eax, dword [ehci_port_number]
    
    .send_reset_signal:
        mov esi, dword [ehci_opbase]
        mov ebx, dword [esi + 44h + eax*4]
        or ebx, 100h
        and ebx, 0FFFFFFFBh
        mov dword [esi + 44h + eax*4], ebx
    
    .pause_1:
        mov ecx, 04000000h
    
    .pause_2:
        sub ecx, 1
        jnz .pause_2
    
    .stop_reset_signal:
        and dword [esi + 44h + eax*4], 0FFFFFEFFh
    
    .pause_3:
        mov ecx, 04000000h
    
    .check_reset:
        test dword [esi + 44h + eax*4], 100h
        jz .return
    
    .pause_4:
        sub ecx, 1
        jnz .check_reset
    
    .return:
        cmp dword [esi + 44h + eax*4], 1005h
        ret

; ==============================================================================

ehci_get_device_descriptor:

    .transport_command:
        mov dword [ehci_TD_buffer_pointer], .request
        mov dword [ehci_TD_buffer_size], 8
        mov dword [ehci_TD_packet_id], 2
        mov dword [ehci_control_toggle], 0
        mov dword [ehci_TD_toggle_pointer], ehci_control_toggle
        mov dword [ehci_TD_endpoint], 0
        mov dword [ehci_TD_bus_address], 0
        call ehci_make_run_TD
    
    .transport_data:
        mov dword [ehci_TD_buffer_pointer], ehci_device_descriptor
        mov dword [ehci_TD_buffer_size], 18
        mov dword [ehci_TD_packet_id], 1
        mov dword [ehci_control_toggle], 1
        mov dword [ehci_TD_toggle_pointer], ehci_control_toggle
        mov dword [ehci_TD_endpoint], 0
        mov dword [ehci_TD_bus_address], 0
        call ehci_make_run_TD
    
    .return:
        ret

    .request:
        db 80h
        db 6
        dw 100h
        dw 0
        dw 18

; ==============================================================================

ehci_get_config_descriptor:
        
        mov eax, dword [ehci_TD_buffer_size]
        mov dword [.total_length], eax
        mov word [.request + 6], ax

    .transport_command:
        mov dword [ehci_TD_buffer_pointer], .request
        mov dword [ehci_TD_buffer_size], 8
        mov dword [ehci_TD_packet_id], 2
        mov dword [ehci_control_toggle], 0
        mov dword [ehci_TD_toggle_pointer], ehci_control_toggle
        mov dword [ehci_TD_endpoint], 0
        mov dword [ehci_TD_bus_address], 0
        call ehci_make_run_TD
    
    .transport_data:
        mov dword [ehci_TD_buffer_pointer], ehci_config_descriptor
        mov eax, dword [.total_length]
        mov dword [ehci_TD_buffer_size], eax
        mov dword [ehci_TD_packet_id], 1
        mov dword [ehci_control_toggle], 1
        mov dword [ehci_TD_toggle_pointer], ehci_control_toggle
        mov dword [ehci_TD_endpoint], 0
        mov dword [ehci_TD_bus_address], 0
        call ehci_make_run_TD
    
    .return:
        ret
    
    .total_length dd 0

    .request:
        db 80h
        db 6
        dw 200h
        dw 0
        dw 9

; ==============================================================================

ehci_set_address:
    
    .prepare:
        mov ax, word [ehci_bus_address]
        mov word [.request + 2], ax

    .transport_command:
        mov dword [ehci_TD_buffer_pointer], .request
        mov dword [ehci_TD_buffer_size], 8
        mov dword [ehci_TD_packet_id], 2
        mov dword [ehci_control_toggle], 0
        mov dword [ehci_TD_toggle_pointer], ehci_control_toggle
        mov dword [ehci_TD_endpoint], 0
        mov dword [ehci_TD_bus_address], 0
        call ehci_make_run_TD
    
    .transport_status:
        mov dword [ehci_TD_buffer_pointer], 0
        mov dword [ehci_TD_buffer_size], 0
        mov dword [ehci_TD_packet_id], 1
        mov dword [ehci_control_toggle], 1
        mov dword [ehci_TD_toggle_pointer], ehci_control_toggle
        mov dword [ehci_TD_endpoint], 0
        mov dword [ehci_TD_bus_address], 0
        call ehci_make_run_TD
        
    .return:
        movzx eax, word [ehci_bus_address]
        mov dword [ehci_TD_bus_address], eax
        ret

    .request:
        db 0
        db 5
        dw 0
        dw 0
        dw 0

; ==============================================================================

ehci_set_config:
    
    .prepare:
        mov al, byte [ehci_config_descriptor + 5]
        mov byte [.request + 2], al

    .transport_command:
        mov dword [ehci_TD_buffer_pointer], .request
        mov dword [ehci_TD_buffer_size], 8
        mov dword [ehci_TD_packet_id], 2
        mov dword [ehci_control_toggle], 0
        mov dword [ehci_TD_toggle_pointer], ehci_control_toggle
        mov dword [ehci_TD_endpoint], 0
        ; ehci_TD_bus_address already set
        call ehci_make_run_TD
        
    .transport_status:
        mov dword [ehci_TD_buffer_pointer], 0
        mov dword [ehci_TD_buffer_size], 0
        mov dword [ehci_TD_packet_id], 1
        mov dword [ehci_control_toggle], 1
        mov dword [ehci_TD_toggle_pointer], ehci_control_toggle
        mov dword [ehci_TD_endpoint], 0
        ; ehci_TD_bus_address already set
        call ehci_make_run_TD
    
    .return:
        ret

    .request:
        db 0
        db 9
        dw 1
        dw 0
        dw 0

; ==============================================================================
 
ehci_inquiry:

    .transport_command:
        mov dword [ehci_TD_buffer_pointer], .request
        mov dword [ehci_TD_buffer_size], 31
        mov dword [ehci_TD_packet_id], 0
        mov dword [ehci_TD_toggle_pointer], ehci_out_toggle
        mov eax, dword [ehci_out_endpoint]
        mov dword [ehci_TD_endpoint], eax
        call ehci_make_run_TD
    
    .pause_1:
        mov ecx, 01000000h
    
    .pause_2:
        sub ecx, 1
        jnz .pause_2
    
    .transport_data:
        mov dword [ehci_TD_buffer_pointer], ehci_inquiry_str
        mov dword [ehci_TD_buffer_size], 36
        mov dword [ehci_TD_packet_id], 1
        mov dword [ehci_TD_toggle_pointer], ehci_in_toggle
        mov eax, dword [ehci_in_endpoint]
        mov dword [ehci_TD_endpoint], eax
        call ehci_make_run_TD
    
    .pause_3:
        mov ecx, 01000000h
    
    .pause_4:
        sub ecx, 1
        jnz .pause_4
    
    .transport_status:
        mov dword [ehci_TD_buffer_pointer], ehci_scsi_csw
        mov dword [ehci_TD_buffer_size], 13
        mov dword [ehci_TD_packet_id], 1
        mov dword [ehci_TD_toggle_pointer], ehci_in_toggle
        mov eax, dword [ehci_in_endpoint]
        mov dword [ehci_TD_endpoint], eax
        call ehci_make_run_TD
    
    .pause_5:
        mov ecx, 01000000h
    
    .pause_6:
        sub ecx, 1
        jnz .pause_6
    
    .return:
        ret

    .request:
        dd 43425355h
        dd 0AABBCCDDh
        dd 36
        db 80h
        db 0
        db 6
        db 12h
        db 0
        db 0
        db 0
        db 36
        db 0
        times 15 db 0
        ;times 10 db 0

; ==============================================================================

ehci_request_sense:

    .transport_command:
        mov dword [ehci_TD_buffer_pointer], .request
        mov dword [ehci_TD_buffer_size], 31
        mov dword [ehci_TD_packet_id], 0
        mov dword [ehci_TD_toggle_pointer], ehci_out_toggle
        mov eax, dword [ehci_out_endpoint]
        mov dword [ehci_TD_endpoint], eax
        call ehci_make_run_TD
    
    .pause_1:
        mov ecx, 01000000h
    
    .pause_2:
        sub ecx, 1
        jnz .pause_2
    
    .transport_data:
        mov dword [ehci_TD_buffer_pointer], ehci_sense
        mov dword [ehci_TD_buffer_size], 18
        mov dword [ehci_TD_packet_id], 1
        mov dword [ehci_TD_toggle_pointer], ehci_in_toggle
        mov eax, dword [ehci_in_endpoint]
        mov dword [ehci_TD_endpoint], eax
        call ehci_make_run_TD
    
    .pause_3:
        mov ecx, 01000000h
    
    .pause_4:
        sub ecx, 1
        jnz .pause_4
    
    .transport_status:
        mov dword [ehci_TD_buffer_pointer], ehci_scsi_csw
        mov dword [ehci_TD_buffer_size], 13
        mov dword [ehci_TD_packet_id], 1
        mov dword [ehci_TD_toggle_pointer], ehci_in_toggle
        mov eax, dword [ehci_in_endpoint]
        mov dword [ehci_TD_endpoint], eax
        call ehci_make_run_TD
    
    .pause_5:
        mov ecx, 01000000h
    
    .pause_6:
        sub ecx, 1
        jnz .pause_6
    
    .return:
        ret

    .request:
        dd 43425355h
        dd 0AABBCCDDh
        dd 18
        db 80h
        db 0
        db 6
        db 03h
        db 0
        dd 0
        db 0
        dw 18
        db 0
        times 6 db 0

; ==============================================================================

ehci_test_unit:

    .transport_command:
        mov dword [ehci_TD_buffer_pointer], .request
        mov dword [ehci_TD_buffer_size], 31
        mov dword [ehci_TD_packet_id], 0
        mov dword [ehci_TD_toggle_pointer], ehci_out_toggle
        mov eax, dword [ehci_out_endpoint]
        mov dword [ehci_TD_endpoint], eax
        call ehci_make_run_TD
    
    .pause_1:
        mov ecx, 01000000h
    
    .pause_2:
        sub ecx, 1
        jnz .pause_2
    
    .transport_status:
        mov dword [ehci_TD_buffer_pointer], ehci_scsi_csw
        mov dword [ehci_TD_buffer_size], 13
        mov dword [ehci_TD_packet_id], 1
        mov dword [ehci_TD_toggle_pointer], ehci_in_toggle
        mov eax, dword [ehci_in_endpoint]
        mov dword [ehci_TD_endpoint], eax
        call ehci_make_run_TD
    
    .pause_3:
        mov ecx, 01000000h
    
    .pause_4:
        sub ecx, 1
        jnz .pause_4
    
    .return:
        ret

    .request:
        dd 43425355h
        dd 0AABBCCDDh
        dd 0
        db 0
        db 0
        db 6
        db 0
        dd 0
        db 0
        db 0
        db 0
        db 0
        times 15 db 0

; ==============================================================================

ehci_read_capacity:

    .transport_command:
        mov dword [ehci_TD_buffer_pointer], .request
        mov dword [ehci_TD_buffer_size], 31
        mov dword [ehci_TD_packet_id], 0
        mov dword [ehci_TD_toggle_pointer], ehci_out_toggle
        mov eax, dword [ehci_out_endpoint]
        mov dword [ehci_TD_endpoint], eax
        call ehci_make_run_TD
    
    .pause_1:
        mov ecx, 01000000h
    
    .pause_2:
        sub ecx, 1
        jnz .pause_2
    
    .transport_data:
        mov dword [ehci_TD_buffer_pointer], ehci_capacity
        mov dword [ehci_TD_buffer_size], 8
        mov dword [ehci_TD_packet_id], 1
        mov dword [ehci_TD_toggle_pointer], ehci_in_toggle
        mov eax, dword [ehci_in_endpoint]
        mov dword [ehci_TD_endpoint], eax
        call ehci_make_run_TD

    .pause_3:
        mov ecx, 01000000h
    
    .pause_4:
        sub ecx, 1
        jnz .pause_4

    .transport_status:
        mov dword [ehci_TD_buffer_pointer], ehci_scsi_csw
        mov dword [ehci_TD_buffer_size], 13
        mov dword [ehci_TD_packet_id], 1
        mov dword [ehci_TD_toggle_pointer], ehci_in_toggle
        mov eax, dword [ehci_in_endpoint]
        mov dword [ehci_TD_endpoint], eax
        call ehci_make_run_TD

    .pause_5:
        mov ecx, 01000000h
    
    .pause_6:
        sub ecx, 1
        jnz .pause_6

    .return:
        ret

    .request:
        dd 43425355h
        dd 0AABBCCDDh
        dd 8
        db 80h
        db 0
        db 10
        db 25h
        db 0
        dd 0
        dw 0
        db 0
        db 0
        times 6 db 0

; ==============================================================================

ehci_read_sectors:

        mov eax, dword [sector_count]
        shl eax, 9
        mov dword [.request + 8], eax
        
        mov eax, dword [lba_sector]
        bswap eax
        mov dword [.request + 17], eax
        
        mov eax, dword [sector_count]
        mov byte [.request + 22], ah
        mov byte [.request + 23], al

    .transport_command:
        mov dword [ehci_TD_buffer_pointer], .request
        mov dword [ehci_TD_buffer_size], 31
        mov dword [ehci_TD_packet_id], 0
        mov dword [ehci_TD_toggle_pointer], ehci_out_toggle
        mov eax, dword [ehci_out_endpoint]
        mov dword [ehci_TD_endpoint], eax
        call ehci_make_run_TD
    
    ;.pause_1:
    ;    mov ecx, 01000000h
    ;
    ;.pause_2:
    ;    sub ecx, 1
    ;    jnz .pause_2
    
    .transport_data:
        mov eax, dword [ram_address]
        mov dword [ehci_TD_buffer_pointer], eax
        mov eax, dword [sector_count]
        shl eax, 9
        mov dword [ehci_TD_buffer_size], eax
        mov dword [ehci_TD_packet_id], 1
        mov dword [ehci_TD_toggle_pointer], ehci_in_toggle
        mov eax, dword [ehci_in_endpoint]
        mov dword [ehci_TD_endpoint], eax
        call ehci_make_run_TD

    ;.pause_3:
    ;    mov ecx, 01000000h
    ;
    ;.pause_4:
    ;    sub ecx, 1
    ;    jnz .pause_4

    .transport_status:
        mov dword [ehci_TD_buffer_pointer], ehci_scsi_csw
        mov dword [ehci_TD_buffer_size], 13
        mov dword [ehci_TD_packet_id], 1
        mov dword [ehci_TD_toggle_pointer], ehci_in_toggle
        mov eax, dword [ehci_in_endpoint]
        mov dword [ehci_TD_endpoint], eax
        call ehci_make_run_TD

    ;.pause_5:
    ;    mov ecx, 01000000h
    ;
    ;.pause_6:
    ;    sub ecx, 1
    ;    jnz .pause_6

    .return:
        ret

    .request:
        dd 43425355h
        dd 0AABBCCDDh
        dd 0
        db 80h
        db 0
        db 10
        db 28h
        db 0
        dd 0
        dw 0
        db 0
        db 0
        times 6 db 0

; ==============================================================================

hidden_ehci_variables:

; -------------------------------------------------------------------
; it is dangerous to have these variables accessible to user programs
; -------------------------------------------------------------------

    ; ehci variables start at 0x00012000
    ehci_bus               equ 0x00012000
    ehci_dev               equ 0x00012004
    ehci_fun               equ 0x00012008
    ehci_port_number       equ 0x0001200C
    ehci_base              equ 0x00012010
    ehci_opbase            equ 0x00012014
    ehci_eecp              equ 0x00012018
    ehci_TD_endpoint       equ 0x00012020
    ehci_TD_bus_address    equ 0x00012024
    ehci_TD_buffer_size    equ 0x00012028
    ehci_TD_buffer_pointer equ 0x0001202C
    ehci_TD_packet_id      equ 0x00012030
    ehci_TD_toggle_pointer equ 0x00012034
    ehci_control_toggle    equ 0x00012040
    ehci_in_toggle         equ 0x00012044
    ehci_out_toggle        equ 0x00012048
    ehci_bus_address       equ 0x00012050
    ehci_in_endpoint       equ 0x00012054
    ehci_out_endpoint      equ 0x00012058
    
    ; control queue head is at 0x0001F000
    ehci_control_QH        equ 0x0001F000
    
    ; device descriptor is received at 0x0001F100
    ehci_device_descriptor equ 0x0001F100
    
    ; configuration descriptor is received at 0x0001F200
    ehci_config_descriptor equ 0x0001F200
    
    ; sense is received at 0x0001F300
    ehci_sense             equ 0x0001F300
    
    ; inquiry string is received at 0x0001F400
    ehci_inquiry_str       equ 0x0001F400
    
    ; capacity is received at 0x0001F500
    ehci_capacity          equ 0x0001F500
    ehci_block_size        equ 0x0001F504
    
    ; scsi csw is received at 0x0001F600
    ehci_scsi_csw          equ 0x0001F600
    
    ; transfer descriptor chain is created starting at 0x00C00000
    ehci_TD_chain          equ 0x00C00000
